/*
 * Copyright (c) 2008, 2009, 2010, 2011, 2013, 2015
 * The Regents of the University of California. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University nor the names of its contributors
 *       may be used to endorse or promote products derived from this software
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char rcsid[] =
    "@(#) $Id: broccoli.c 1065 2015-02-19 01:50:47Z leres $ (LBL)";
#endif

#include <sys/types.h>
#include <sys/time.h>

#include <arpa/inet.h>

#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <unistd.h>

#include "acld.h"
#include "broccoli.h"

/* Locals */
static const char *requestevstr = EV_REQUEST_STR;
static const char *replyevstr = EV_REPLY_STR;
static struct state *statep;

/* Forwards */
void brocallback(BroConn *, void *, BroRecord *);
int broextractstr(BroRecord *, const char *, char **);
int broextractuint32(BroRecord *, const char *, uint32 *);

/* Broccoli version of clientinput() */
void
brocallback(BroConn *bc, void *data, BroRecord *rec)
{
	int i;
	size_t len;
	struct client *cl;
	char *cmd, *cmd2, *cp, *cp2;
	struct req *rp;

	cl = data;
	rp = cl->treq;
	if (rp != NULL) {
		lg(LOG_ERR, "brocallback: cl->treq not NULL");
		exit(EX_SOFTWARE);
	}

	rp = newreq("brocallback req");
	rp->state = RSTATE_PENDING;

	/* Arrival timestamp */
	getts(&rp->ats);

	/* Command string */
	if (!broextractstr(rec, EV_CMD, &cmd)) {
		clientsenderr(cl, rp, "Missing \"cmd\"");
		freereq(statep, rp);
		return;
	}

	/* Cookie */
	if (!broextractuint32(rec, EV_COOKIE, &rp->cookie)) {
		clientsenderr(cl, rp, "Missing \"cookie\"");
		freereq(statep, rp);
		return;
	}

	/* Insert cookie as the 2nd argument as expected by clientfinishreq() */
	len = strlen(cmd) + 16, sizeof(*cmd2);
	cmd2 = new(len, sizeof(*cmd2), "brocallback");

	cp = cmd;
	cp2 = cmd2;
	while (len > 0 && !isspace(*cp) && *cp != '\0') {
		*cp2++ = *cp++;
		--len;
	}
	i = snprintf(cp2, len, " %u", rp->cookie);
	cp2 += i;
	len -= i;
	strncpy(cp2, cp, len - 1);
	cp2[len - 1] = '\0';

	rp->an = makeargv(cmd2, &rp->av);
	free(cmd);
	free(cmd2);

	/* Optional comment string */
	if (broextractstr(rec, EV_COMMENT, &cp))
		rp->comment = cp;

	if (debug) {
		fprintf(log_lf.f, "%s brocallback: client #%u: \"%s",
		    tsstr(), cl->n, rp->an > 0 ? rp->av[0] : "");
		for (i = 1; i < rp->an; ++i)
			fprintf(log_lf.f, " %s", rp->av[i]);
		fprintf(log_lf.f, "\"");
		if (rp->comment != NULL)
			fprintf(log_lf.f, " comment \"%s\"", rp->comment);
		fprintf(log_lf.f, "\n");
	}

	clientfinishreq(getstate(), cl, rp);
}

int
broextractstr(BroRecord *rec, const char *name, char **strp)
{
	int btype;
	BroString *bstr;

	btype = BRO_TYPE_STRING;
	bstr = bro_record_get_named_val(rec, name, &btype);
	if (bstr == NULL)
		return (0);
	*strp = new(1, bstr->str_len + 1, "broextractstr");
	strncpy(*strp, (const char *)bstr->str_val, bstr->str_len);
	(*strp)[bstr->str_len] = '\0';
	bro_string_cleanup(bstr);
	return (1);
}

int
broextractuint32(BroRecord *rec, const char *name, uint32 *up)
{
	int btype;
	uint32 *bup;

	btype = BRO_TYPE_COUNT;
	bup = bro_record_get_named_val(rec, name, &btype);
	if (bup == NULL) {
		*up = 0;
		return (0);
	}
	*up = *bup;
	return (1);
}

/* Returns true if successful else false */
int
broinit(struct state *sp, struct client *cl)
{
	cl->bc = bro_conn_new_socket(cl->c, BRO_CFLAG_NONE);
	if (cl->bc == NULL) {
		lg(LOG_DEBUG, "client #%d bro_conn_new_socket failed", cl->n);
		return (0);
	}

	/* Set the connection class */
	bro_conn_set_class(cl->bc, "acld");

	bro_event_registry_add(cl->bc, requestevstr,
	    (BroEventFunc)brocallback, cl);
	if (!bro_conn_connect(cl->bc)) {
		lg(LOG_DEBUG, "client #%d bro_conn_connect failed", cl->n);
		return (0);
	}
	statep = sp;
	return (1);
}

/* Returns true if successful else false (callback side effect) */
int
broinput(struct client *cl)
{

	(void)bro_conn_process_input(cl->bc);
	return (bro_conn_alive(cl->bc));
}

void
brosend(struct client *cl, struct req *rp)
{
	int success;
	double time;
	BroEvent *ev;
	BroRecord *rec;
	BroString str;

	lg(LOG_DEBUG, "%s: #%d send", prog, cl->n);

	rec = bro_record_new();
	if (rec == NULL) {
		lg(LOG_ERR, "bro_record_new() failed");
		return;
	}

	/* "cmd" */
	bro_string_init(&str);
	bro_string_set(&str, rp->cmd);
	if (!bro_record_add_val(rec, EV_CMD, BRO_TYPE_STRING, NULL, &str)) {
		lg(LOG_ERR, "bro_record_add_val(%s) failed", EV_CMD);
		bro_record_free(rec);
		return;
	}
	bro_string_cleanup(&str);

	/* cookie */
	if (!bro_record_add_val(rec, EV_COOKIE, BRO_TYPE_COUNT,
	    NULL, &rp->cookie)) {
		lg(LOG_ERR, "bro_record_add_val(%s) failed", EV_COOKIE);
		bro_record_free(rec);
		return;
	}

	/* timestamp */
	time = bro_util_timeval_to_double(&rp->cts);
	if (!bro_record_add_val(rec, EV_TIME, BRO_TYPE_TIME, NULL, &time)) {
		lg(LOG_ERR, "bro_record_add_val(%s) failed", EV_TIME);
		bro_record_free(rec);
		return;
	}

	/* success */
	success = ((rp->flags & RFLAG_FAILED) == 0);
	if (!bro_record_add_val(rec, EV_SUCCESS, BRO_TYPE_BOOL,
	    NULL, &success)) {
		lg(LOG_ERR, "bro_record_add_val(%s) failed", EV_SUCCESS);
		bro_record_free(rec);
		return;
	}

	if (IOBUF_LEN(&rp->payload) > 0) {
		bro_string_init(&str);
		bro_string_set(&str, rp->payload.buf);
		if (!bro_record_add_val(rec, EV_PAYLOAD, BRO_TYPE_STRING,
		    NULL, &str)) {
			lg(LOG_ERR, "bro_record_add_val(%s) failed",
			    EV_PAYLOAD);
			bro_record_free(rec);
			return;
		}
		bro_string_cleanup(&str);
	}

	/* Build event */
	ev = bro_event_new(replyevstr);

	bro_event_add_val(ev, BRO_TYPE_RECORD, NULL, rec);
	bro_record_free(rec);

	if (!bro_event_send(cl->bc, ev))
		lg(LOG_DEBUG, "%s: #%d failed to send", prog, cl->n);
	else
		lg(LOG_DEBUG, "%s: #%d sent", prog, cl->n);
	bro_event_free(ev);
	rp->state = RSTATE_DONE;
}
