/*
 * Copyright (c) 2012, 2013, 2014, 2015
 * The Regents of the University of California. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University nor the names of its contributors
 *       may be used to endorse or promote products derived from this software
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char rcsid[] =
    "@(#) $Id: junoscript.c 1065 2015-02-19 01:50:47Z leres $ (LBL)";
#endif

#include <sys/types.h>
#include <sys/mman.h>
#include <sys/time.h>

#include <arpa/inet.h>

#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <sysexits.h>
#include <unistd.h>

#include <libssh2.h>

#include "acld.h"
#include "child.h"
#include "junoscript.h"
#include "rpc.h"
#include "timer.h"

/* Locals */
static const char suffix_host[] = "HOST";
static const char suffix_hostport[] = "HOSTPORT";
static const char suffix_port[] = "PORT";
static struct log log_js;
static char remote[64];			/* Name of router we're connecting to */

static const char *aclsuffixes[] = {
	suffix_host,
	suffix_hostport,
	suffix_port,
	NULL,
};
#define NUMACLSUFFIXES ((sizeof(aclsuffixes) / sizeof(*aclsuffixes)) - 1)

/* Points to next ACL suffix to list */
static const char **nextaclsuffix;

static struct s2v acl2aclsuffix[] = {
	{ suffix_host,		ATYPE_BLOCKHOST },
	{ suffix_host,		ATYPE_BLOCKNET },
	{ suffix_hostport,	ATYPE_PERMITUDPDSTHOSTPORT },
	{ suffix_hostport,	ATYPE_PERMITTCPDSTHOSTPORT },
	{ suffix_port,		ATYPE_BLOCKUDPPORT },
	{ suffix_port,		ATYPE_BLOCKTCPPORT },
	{ NULL,			-1, }
};

static LIBSSH2_SESSION *session;
static LIBSSH2_CHANNEL *channel;

static const char js_vers[] = "1.0";
static const char js_rel[] = "12.1R1.9";

static const char js_pi_fmt[64] =
    "<?xml version=\"%s\" encoding=\"us-ascii\"?>";
static const char js_js_fmt[64] =
    "<junoscript version=\"%s\" hostname=\"%s\" release=\"%s\">";

static const char js_abort[] = "<abort/>";
static char js_abort_acknowledgement[] = "<abort-acknowledgement/>\n";

static const char xml_comment[] = "<!--";
static const char xml_junoscript[] = "<junoscript ";
static const char xml_xml[] = "<?xml ";

/* Arguments: family, aclname, family */
static const char js_listacl_fmt[] =
    "<rpc>"
	 "<get-configuration>"
	    "<configuration>"
		"<firewall>"
		    "<family>"
			"<%s>"
			    "<filter>"
				"<name>%s</name>"
			    "</filter>"
			"</%s>"
		    "</family>"
		"</firewall>"
	    "</configuration>"
	"</get-configuration>"
    "</rpc>";

/* Arguments: family, aclname, seq, addr, family */
static const char js_drop_fmt[] =
    "<rpc>"
	 "<load-configuration action=\"merge\">"
	    "<configuration>"
		"<firewall>"
		    "<family>"
			"<%s>"
			    "<filter>"
				"<name>%s</name>"
				"<term>"
				    "<name>%d</name>"
				    "<from>"
					"<address>"
					    "<name>%s</name>"
					"</address>"
				    "</from>"
				    "<then>"
					"<discard>"
					"</discard>"
				    "</then>"
				"</term>"
			    "</filter>"
			"</%s>"
		    "</family>"
		"</firewall>"
	    "</configuration>"
	"</load-configuration>"
    "</rpc>";

/* Arguments: aclname, seq, protocol, port */
static const char js_dropport4_fmt[] =
    "<rpc>"
	 "<load-configuration action=\"merge\">"
	    "<configuration>"
		"<firewall>"
		    "<family>"
			"<inet>"
			    "<filter>"
				"<name>%s</name>"
				"<term>"
				    "<name>%d</name>"
				    "<from>"
					"<protocol>%s</protocol>"
					"<destination-port>"
					    "%s"
					"</destination-port>"
				    "</from>"
				    "<then>"
					"<discard>"
					"</discard>"
				    "</then>"
				"</term>"
			    "</filter>"
			"</inet>"
		    "</family>"
		"</firewall>"
	    "</configuration>"
	"</load-configuration>"
    "</rpc>";

/* Arguments: aclname, seq, protocol, port */
static const char js_dropport6_fmt[] =
    "<rpc>"
	 "<load-configuration action=\"merge\">"
	    "<configuration>"
		"<firewall>"
		    "<family>"
			"<inet6>"
			    "<filter>"
				"<name>%s</name>"
				"<term>"
				    "<name>%d</name>"
				    "<from>"
					"<next-header>%s</next-header>"
					"<destination-port>"
					    "%s"
					"</destination-port>"
				    "</from>"
				    "<then>"
					"<discard>"
					"</discard>"
				    "</then>"
				"</term>"
			    "</filter>"
			"</inet6>"
		    "</family>"
		"</firewall>"
	    "</configuration>"
	"</load-configuration>"
    "</rpc>";

/* Arguments: aclname, seq, addr, protocol, port */
static const char js_permithostport4_fmt[] =
    "<rpc>"
	 "<load-configuration action=\"merge\">"
	    "<configuration>"
		"<firewall>"
		    "<family>"
			"<inet>"
			    "<filter>"
				"<name>%s</name>"
				"<term>"
				    "<name>%d</name>"
				    "<from>"
					"<destination-address>"
					    "<name>%s</name>"
					"</destination-address>"
					"<protocol>%s</protocol>"
					"<destination-port>"
					    "%s"
					"</destination-port>"
				    "</from>"
				    "<then>"
					"<accept/>"
				    "</then>"
				"</term>"
			    "</filter>"
			"</inet>"
		    "</family>"
		"</firewall>"
	    "</configuration>"
	"</load-configuration>"
    "</rpc>";

/* Arguments: aclname, seq, addr, protocol, port */
static const char js_permithostport6_fmt[] =
    "<rpc>"
	 "<load-configuration action=\"merge\">"
	    "<configuration>"
		"<firewall>"
		    "<family>"
			"<inet6>"
			    "<filter>"
				"<name>%s</name>"
				"<term>"
				    "<name>%d</name>"
				    "<from>"
					"<destination-address>"
					    "<name>%s</name>"
					"</destination-address>"
					"<next-header>%s</next-header>"
					"<destination-port>"
					    "%s"
					"</destination-port>"
				    "</from>"
				    "<then>"
					"<accept/>"
				    "</then>"
				"</term>"
			    "</filter>"
			"</inet6>"
		    "</family>"
		"</firewall>"
	    "</configuration>"
	"</load-configuration>"
    "</rpc>";

/* Arguments: family, aclname, seq, family */
static const char js_delete_term_fmt[] =
    "<rpc>"
	 "<load-configuration action=\"merge\">"
	    "<configuration>"
		"<firewall>"
		    "<family>"
			"<%s>"
			    "<filter>"
				"<name>%s</name>"
				"<term delete=\"delete\">"
				    "<name>%d</name>"
				"</term>"
			    "</filter>"
			"</%s>"
		    "</family>"
		"</firewall>"
	    "</configuration>"
	"</load-configuration>"
    "</rpc>";

static const char js_rename_start[] =
    "<rpc>"
	 "<load-configuration action=\"merge\">"
	    "<configuration>"
		"<firewall>";

/* Arguments: family, aclname */
static const char js_filter_start_fmt[] =
		    "<family>"
			"<%s>"
			    "<filter>"
				"<name>%s</name>";

/* Arguments: newseq, oldseq */
static const char js_rename_fmt[] =
				"<term rename=\"rename\" name=\"%d\">"
				    "<name>%d</name>"
				"</term>";

/* Arguments: family */
static const char js_filter_end_fmt[] =
			    "</filter>"
			"</%s>"
		    "</family>";

static const char js_rename_end[] =
		"</firewall>"
	    "</configuration>"
	"</load-configuration>"
    "</rpc>";

/* Arguments: comment */
static const char js_commit_fmt[] =
    "<rpc>"
	 "<commit-configuration>"
	     "<synchronize/>"
	     "<log>%s</log>"
	 "</commit-configuration>"
    "</rpc>";

static const char js_open_private_configuration[] =
    "<rpc>"
	"<open-configuration>"
	    "<private/>"
	"</open-configuration>"
    "</rpc>";

static const char js_discard_private_configuration[] =
    "<rpc>"
	"<close-configuration/>"
    "</rpc>";

static const char js_uptime_cmd[] =
    "<rpc>"
	"<get-system-uptime-information/>"
    "</rpc>";

static const char js_request_end_session[] =
    "<rpc>"
	"<request-end-session/>"
    "</rpc>";

/* Forwards */
static const char *family2str(u_int);
static void junoscriptchildlistroute(struct state *);
static void junoscriptclientcompact(struct state *, struct client *,
    struct req *);
static void junoscriptconnect(struct state *);
static void junoscriptflush(struct state *);
static int junoscriptinput(struct state *, struct client *);
static void junoscriptkill(struct state *);
static void junoscriptlistacl(struct state *);
static void junoscriptlistroute(struct state *, struct client *,
    struct req *, int);
static void junoscriptlog(const char *, const char *);
static void junoscriptlogin(struct state *);
static const char *junoscriptmapaclname(const char *, struct acl *);
static void junoscriptnullzero(struct state *, struct client *, struct req *);
static void __attribute__ ((format (printf, 2, 3)))
    junoscriptqueue(struct state *, const char *, ...);
static void junoscriptquerynullzero(struct state *, struct client *,
    struct req *);
static int junoscriptread(int, struct iobuf *);
static void junoscriptrequest(struct state *, struct aclgroup *, struct req *);
static void __attribute__ ((format (printf, 2, 3)))
    junoscriptsend(struct state *, const char *, ...);
static void junoscriptsendattr(struct state *);
static void junoscriptserverayt(struct state *);
static void junoscriptservercompact(struct state *);
static void junoscriptserversync(struct state *);

static const char *
family2str(u_int family)
{
	switch (family) {

	case AF_INET:
		return ("inet");

	case AF_INET6:
		return ("inet6");

	default:
		abort();
	}
}

static const char *
junoscriptcheckreq(struct req *rp)
{
	const char *stype, *sreq;
	static char buf[132];

	switch (rp->type) {

	case REQ_DROP:
	case REQ_DROPUDPPORT:
	case REQ_DROPTCPPORT:
	case REQ_PERMITUDPDSTHOSTPORT:
	case REQ_PERMITTCPDSTHOSTPORT:
	case REQ_RESTORE:
	case REQ_RESTOREUDPPORT:
	case REQ_RESTORETCPPORT:
	case REQ_UNPERMITUDPDSTHOSTPORT:
	case REQ_UNPERMITTCPDSTHOSTPORT:
		sreq = NULL;
		break;

	default:
		sreq = val2str(cmd2req, rp->type);
		break;
	}

	switch (rp->acl.type) {

	case ATYPE_BLOCKHOST:
	case ATYPE_BLOCKNET:
	case ATYPE_BLOCKUDPPORT:
	case ATYPE_BLOCKTCPPORT:
	case ATYPE_PERMITUDPDSTHOSTPORT:
	case ATYPE_PERMITTCPDSTHOSTPORT:
		stype = NULL;
		break;

	default:
		stype = val2str(str2acl, rp->acl.type);
		break;
	}
	if (stype == NULL && sreq == NULL)
		return (NULL);

	if (sreq == NULL)
		sreq = val2str(cmd2req, rp->type);
	if (stype == NULL)
		stype = val2str(str2acl, rp->acl.type);

	if (strcmp(sreq, stype) == 0)
		(void)snprintf(buf, sizeof(buf),
		    "junoscript request %s unsupported", sreq);
	else
		(void)snprintf(buf, sizeof(buf),
		    "junoscript request %s unsupported (type %s)", sreq, stype);
	return (buf);
}

static void
junoscriptchildlistroute(struct state *sp)
{
	lg(LOG_DEBUG, "junoscriptchildlistroute: not implemented");
	++sp->listedroutes;
}

static void
junoscriptclientcompact(struct state *sp, struct client *cl, struct req *rp)
{
	int i, j, seq, n, didstart, didrename;
	int *vp;
	const char *p;
	const char *sfamily;
	struct aclgroup *gp;
	struct acllist *ap;
	struct cf *cf;
	struct acl *al;
	struct array *dp, *dp2;

	cf = sp->cf;
	gp = aclfindgroupbyname(cf, rp->aclname);
	if (gp == NULL) {
		clientsenderr(cl, rp, "ACL %s unknown", rp->aclname);
		return;
	}
	gp->lastseq = 0;
	gp->lastportseq = 0;
	gp->lastpermithostportseq = 0;
	sfamily = family2str(ACLGROUP_FAMILY(gp));
	didstart = 0;
	dp = &gp->acllist_array;
	for (i = 0, ap = gp->acllist; i < dp->len; ++i, ++ap) {
		/* Bail if there's nothing to do */
		dp2 = &ap->acl_array;
		if (dp2->len == 0)
			continue;

		/* Update lastseq, lastportseq and lastpermithostportseq */
		p = val2str(acl2aclsuffix, ap->acl->type);
		if (p == suffix_host) {
			vp = &gp->lastseq;
			n = cf->c_lowseq;
		} else if (p == suffix_hostport) {
			vp = &gp->lastpermithostportseq;
			n = cf->c_lowpermithostportseq;
		} else if (p == suffix_port) {
			vp = &gp->lastportseq;
			n = cf->c_lowportseq;
		} else
			abort();
		seq = 0;
		didrename = 0;
		for (j = 0, al = ap->acl; j < dp2->len; ++j, ++al) {
			seq = n + j;
			if (al->seq == seq)
				continue;
			if (!didstart) {
				/* Start configuring */
				junoscriptqueue(sp, js_rename_start);
				++didstart;
			}
			if (!didrename) {
				/* Start filter */
				junoscriptqueue(sp, js_filter_start_fmt,
				    sfamily, ap->name);
				++didrename;
			}
			/* Rename one term */
			junoscriptqueue(sp, js_rename_fmt, seq, al->seq);
			al->seq = seq;
		}
		*vp = seq;
		if (didrename) {
			/* Finish filter */
			junoscriptqueue(sp, js_filter_end_fmt, sfamily);
		}
	}
	if (!didstart) {
		clientsendfmt(cl, rp, "ACL %s already compact", rp->aclname);
		return;
	}

	/* Finish configuring */
	junoscriptqueue(sp, js_rename_end);

	/* Update request state */
	rp->state = RSTATE_CHILD;

	/* XXX if the commit fails we will need to re-read these ACLs */
}

void
junoscriptcommit(struct state *sp)
{
	char buf[132];

	(void)snprintf(buf, sizeof(buf), "%s@%s", sp->cf->c_euser, hostname);
	junoscriptsend(sp, js_commit_fmt, buf);
}

static void
junoscriptconnect(struct state *sp)
{
	int s, rc, opterrno;
	socklen_t len;
	const char *p;
	struct cf *cf;

	/* Always disable the child connect hook */
	sp->f_childconnect = NULL;

	/* Make sure we still  have a socket */
	s = sp->wfd;
	if (s < 0)
		return;

	/* We're committed now */
	sp->wfd = -1;

	/* Status check on the socket */
	len = sizeof(opterrno);
	if (getsockopt(s, SOL_SOCKET, SO_ERROR, &opterrno, &len) < 0) {
		lg(LOG_ERR, "junoscriptconnect: %s: getsockopt: %s",
		    remote, strerror(errno));

		/* Clean up */
		(void)close(s);
		return;
	}
	if (opterrno != 0) {
		lg(LOG_ERR, "connect error to %s: %s",
		    remote, strerror(opterrno));

		/* Clean up */
		(void)close(s);
		return;
	}

	/* Turn blocking I/O back on */
	blocking(s, 1);

	/* Open a channel */
	session = libssh2_session_init();
	if (libssh2_session_handshake(session, s) < 0) {
		lg(LOG_ERR,
		    "junoscriptconnect: %s: libssh2_session_handshake failed",
		    remote);

		/* Clean up */
		(void)close(s);
		return;
	}

#ifdef notdef
	/* XXX todo: check the router's ssh fingerprint (SSHFP?) */
	p = libssh2_hostkey_hash(session, LIBSSH2_HOSTKEY_HASH_SHA1);
#endif
	cf = sp->cf;
	rc = libssh2_userauth_publickey_fromfile(session, cf->c_euser,
	    cf->c_sshpubfn, cf->c_sshkeyfn, NULL);
	if (rc < 0) {
		lg(LOG_ERR, "libssh2_userauth_publickey_fromfile: %d", rc);
		switch (errno) {

		case EAGAIN:
			lg(LOG_ERR, "connect: %s: Invalid ssh credentials",
			    remote);
			break;

		case ENOENT:
			lg(LOG_ERR, "connect: %s: ssh credentials missing",
			    remote);
			break;

		default:
			lg(LOG_ERR, "junoscriptconnect: %s:"
			    " libssh2_userauth_publickey_fromfile failed: %s",
			    remote, strerror(errno));
			exit(EX_SOFTWARE);
		}

		/* Clean up */
		libssh2_session_free(session);
		session = NULL;
		(void)close(s);
		return;
	}

	/* Open a channel */
	channel = libssh2_channel_open_session(session);
	if (channel == NULL) {
		lg(LOG_ERR,
		    "junoscriptconnect: %s: failed to open a channel", remote);

		/* Clean up */
		libssh2_session_free(session);
		session = NULL;
		(void)close(s);
		return;
	}

	p = libssh2_session_banner_get(session);
	if (p == NULL) {
		lg(LOG_ERR, "junoscriptconnect:"
		    " %s: libssh2_session_banner_get failed", remote);

		/* Clean up */
		(void)close(s);
		libssh2_channel_free(channel);
		channel = NULL;
		libssh2_session_free(session);
		session = NULL;
		return;
	}

	/* We're connected now */
	lg(LOG_DEBUG, "connected to %s: %s", remote, p);
	setastate(sp, ASTATE_CONNECTED);

	/* Get a channel */
	for (;;) {
		rc = libssh2_channel_exec(channel, "junoscript");
		if (rc == LIBSSH2_ERROR_EAGAIN) {
			lg(LOG_ERR, "junoscriptconnect: Would block; sleeping");
			sleep(1);
			continue;
		}
		break;
	}
	if (rc < 0) {
		lg(LOG_ERR, "junoscriptconnect:"
		    " %s: libssh2_channel_exec failed: %d", remote, rc);

		/* Clean up */
		(void)close(s);
		libssh2_channel_free(channel);
		channel = NULL;
		libssh2_session_free(session);
		session = NULL;
		setastate(sp, ASTATE_NOTCONNECTED);
		return;
	}

	/* Configure non-blocking */
	libssh2_session_set_blocking(session, 0);

	sp->rfd = s;
	sp->wfd = s;

	/* Emit PI */
	junoscriptsend(sp, js_pi_fmt, js_vers);

	/* Emit tag */
	junoscriptsend(sp, js_js_fmt, js_vers, hostname, js_rel);

	/* Reset login timer */
	timerreset(&sp->t_login);
}

void
junoscriptexitprivate(struct state *sp)
{
	if (sp->sentprivate) {
		junoscriptsend(sp, js_discard_private_configuration);
		sp->sentprivate = 0;
	}
}

static void
junoscriptflush(struct state *sp)
{
	int rc;
	size_t n;
	struct iobuf *ip;

	ip = &sp->wbuf;
	n = IOBUF_LEN(ip);
	if (n > 0) {
		rc = libssh2_channel_write(channel, ip->buf, n);
		if (rc == LIBSSH2_ERROR_EAGAIN) {
			lg(LOG_ERR, "libssh2_channel_write: eagain");
			return;
		}
		if (rc < 0) {
			/* XXX kill off child here? */
			lg(LOG_ERR, "libssh2_channel_write: %s",
			    strerror(errno));
			return;
		}

		if (libssh2_channel_eof(channel)) {
			/* XXX kill off child here? */
			lg(LOG_ERR, "junoscriptflush: EOF");
			return;
		}

		/* Log the line */
		ip->buf[n] = '\0';
		junoscriptlog("send", ip->buf);
		IOBUF_LEN(ip) = 0;
	}
}

void
junoscriptinit(struct state *sp)
{
	int i, j;
	const char **pp;
	struct cf *cf;
	struct aclgroup *gp;
	struct acllist *ap;
	struct array *dp, *dp2, *dp3;
	char aclname[64];

	/* Static initialization */
	nextaclsuffix = NULL;
	sp->sentprivate = 0;
	IOBUF_INIT(&sp->wbuf);

	/* Child hooks */
	sp->f_childcheckreq = junoscriptcheckreq;
	sp->f_childconnect = NULL;
	sp->f_childinput = junoscriptinput;
	sp->f_childkill = junoscriptkill;
	sp->f_childlistacl = junoscriptlistacl;
	sp->f_childlistroute = junoscriptchildlistroute;
	sp->f_childlogin = junoscriptlogin;
	sp->f_childread = junoscriptread;
	sp->f_childrequest = junoscriptrequest;
	sp->f_childsend = junoscriptsend;
	sp->f_childsendattr = junoscriptsendattr;

	/* Client hooks */
	sp->f_clientcompact = junoscriptclientcompact;
	sp->f_clientlistroute = junoscriptlistroute;
	sp->f_clientmapaclname = junoscriptmapaclname;
	sp->f_clientnullzero = junoscriptnullzero;
	sp->f_clientquerynullzero = junoscriptquerynullzero;

	/* Server hooks */
	sp->f_serverayt = junoscriptserverayt;
	sp->f_servercompact = junoscriptservercompact;
	sp->f_serversync = junoscriptserversync;

	/* Junoscript log */
	if (sp->cf->c_junoscriptlog != NULL) {
		log_js.fn = sp->cf->c_junoscriptlog;
		checklog(&log_js);
	}

	/* Clobber remote server */
	strncpy(remote, "?", sizeof(remote));
	remote[sizeof(remote) - 1] = '\0';

	/* Check API for match with the DLL we are using */
	LIBXML_TEST_VERSION;

	/* Expand each ACL group to have the necessary number of ACL lists */
	cf = sp->cf;
	dp = &cf->c_aclgroup_array;
	for (i = 0, gp = cf->c_aclgroup; i < dp->len; ++i, ++gp) {
		dp2 = &gp->acllist_array;
		DYNARRAY(dp2, gp->acllist, NUMACLSUFFIXES,
		    "junoscriptinit acllist");
		dp2->len = NUMACLSUFFIXES;
		free(gp->acllist->name);
		gp->acllist->name = NULL;
		pp = aclsuffixes;
		for (j = 0, ap = gp->acllist; j < dp2->len; ++j, ++ap) {
			(void)snprintf(aclname, sizeof(aclname), "%s-%s",
			    gp->name, *pp);
			++pp;
			ap->name = strsave(aclname);

			dp3 = &ap->acl_array;
			dp3->osize = sizeof(*ap->acl);
			dp3->inclen = gp->acllist->acl_array.inclen;
		}
	}
}

/* Return 1 if done else 0 */
static int
junoscriptinput(struct state *sp, struct client *cl)
{
	int i;
	size_t cc;
	char *cp;
	struct cf *cf;
	struct iobuf *ip;
	struct req *rp;
	xmlDocPtr doc;
	xmlNode *root;
	const char *errmsg;
	char buf[128];

	cf = sp->cf;
	ip = &sp->rbuf;
	if (IOBUF_LEN(ip) == 0)
		return (1);

	switch (sp->state) {

	case ASTATE_CONNECTED:
		/*
		 * Look for the comment that says we're logged in:
		 *
		 *    <!-- user zeek, class j-superuser-enforce-private -->\n
		 */
		while (iohaveline(ip) && (cp = iogetstr(ip)) != NULL) {
			if (ISXMLCOMMENT(cp)) {
				if (strstr(cp, "session start") != NULL ||
				    strstr(cp, "No zombies") != NULL)
					continue;
				(void)snprintf(buf, sizeof(buf),
				    "<!-- user %s,", cf->c_euser);
				if (strncmp(cp, buf, strlen(buf)) == 0) {
					setastate(sp, ASTATE_LOGGEDIN);
					timerreset(&sp->t_login);
					break;
				}
				if (debug)
					lg(LOG_DEBUG, "junoscriptinput: \"%s\"",
					    pretty(cp));
				continue;
			}
			if (ISXMLTAG(cp, xml_xml) ||
			    ISXMLTAG(cp, xml_junoscript)) {
				if (debug > 1)
					lg(LOG_DEBUG, "junoscriptinput: \"%s\"",
					    pretty(cp));
				continue;
			}
			lg(LOG_DEBUG, "junoscriptinput: unexpected \"%s\"",
				pretty(cp));
		}
		break;

	case ASTATE_SENTLISTACL:
	case ASTATE_LOGGEDIN:
		break;

	default:
		lg(LOG_DEBUG, "junoscriptinput: %s not implemented",
		    val2str(astate2str, sp->state));
		abort();
	}

	while (iohaveline(ip) && ISXMLCOMMENT(ip->buf) &&
	    (cp = iogetstr(ip)) != NULL) {
		/* Did the router time our session out? */
		if (strstr(cp, "session end") != NULL) {
			lg(LOG_ERR, "junoscriptinput: router ended session");
			setastate(sp, ASTATE_NOTCONNECTED);
			junoscriptkill(sp);
			return (1);
		}

		if (debug)
			lg(LOG_DEBUG, "junoscriptinput: # %s", pretty(cp));
	}

	/* Look for an abort acknowledgement */
	cc = sizeof(js_abort_acknowledgement) - 1;
	if (IOBUF_LEN(ip) >= cc &&
	    strncmp(ip->buf, js_abort_acknowledgement, cc) == 0) {
		lg(LOG_INFO, "abort acknowled");
		ioeat(ip, cc);
	}

	cc = iohaverpcreply(ip);
	if (cc == 0)
		return (1);
	timerset(&sp->t_ayt, cf->c_ayt_secs);
	doc = xmlReadMemory(ip->buf, cc, "j.xml", NULL, XML_PARSE_NOERROR);
	ioeat(ip, cc);
	if (doc == NULL) {
		i = cc;
		if (i > 16)
			i = 16;
		lg(LOG_DEBUG,
		    "junoscriptinput: xmlReadMemory failed: \"%.*s ...\"",
		    i, ip->buf);
		return (0);
	}
	if (cl != NULL)
		rp = cl->req;
	else
		rp = sp->req;

	/* Parse the XML */
	root = xmlDocGetRootElement(doc);
	errmsg = NULL;
	switch (sp->state) {

	case ASTATE_SENTLISTACL:
		errmsg = rpc_process(sp, cl, rp, root);
		setastate(sp, ASTATE_LOGGEDIN);
		timerreset(&sp->t_login);
		break;

	case ASTATE_LOGGEDIN:
		errmsg = rpc_process(sp, cl, rp, root);
		break;

	default:
		lg(LOG_ERR, "junoscriptinput: Can't handle state %d",
		    sp->state);
		break;
	}

	if (errmsg != NULL)
		lg(LOG_DEBUG, "junoscriptinput: %s", pretty(errmsg));

	xmlFreeDoc(doc);
	xmlCleanupParser();

	/*
	 * Fix up things if the router command failed.
	 *
	 * If we requested a private copy of the configuration,
	 * have pending configuration changes to send and got a
	 * reply without an error, send the changes.
	 */
	ip = &sp->wbuf;
	if (rp != NULL) {
		if ((rp->flags & RFLAG_FAILED) != 0)
			childaclcleanup(sp, rp);
		else if (sp->sentprivate && rp->state == RSTATE_CHILD) {
			if (IOBUF_LEN(ip) > 0)
				junoscriptflush(sp);
		}
	}
#ifdef notdef
	if (sp->sentprivate) {
		sp->sentprivate = 0;
		// IOBUF_LEN(ip) = 0;
		if (rp != NULL && rp->state == RSTATE_CHILD)
			clientsenderr(cl, rp,
			    "junoscriptinput: unexpected state");
	}
#endif

	return (0);
}

static void
junoscriptkill(struct state *sp)
{
	if (sp->wfd >= 0 && sp->state == ASTATE_LOGGEDIN)
		junoscriptsend(sp, js_request_end_session);
	setastate(sp, ASTATE_NOTCONNECTED);
	sp->sentprivate = 0;
	sp->cmd = NULL;
	IOBUF_INIT(&sp->wbuf);

	(void)libssh2_session_disconnect(session, "goodbye");

	if (channel != NULL) {
		libssh2_channel_free(channel);
		channel = NULL;
	}
	if (session != NULL) {
		libssh2_session_free(session);
		session = NULL;
	}

	if (sp->rfd >= 0) {
		(void)close(sp->rfd);
		/* In case they are the same fd (e.g. libssh2) */
		if (sp->rfd == sp->wfd)
			sp->wfd = -1;
		sp->rfd = -1;
	}

	if (sp->wfd >= 0) {
		(void)close(sp->wfd);
		sp->wfd = -1;
	}

	iofree(&sp->rbuf);

	timerreset(&sp->t_ayt);
	timerreset(&sp->t_login);
	timerreset(&sp->t_sync);
}

/* Loop over the junos ACL suffix types */
static void
junoscriptlistacl(struct state *sp)
{
	int i;
	const char *sfamily;
	struct cf *cf;
	struct aclgroup *gp;
	struct array *dp;
	char aclname[64];

	cf = sp->cf;
	dp = &cf->c_aclgroup_array;
	for (i = 0, gp = cf->c_aclgroup; i < dp->len; ++i, ++gp)
		if (!gp->sentlistacl) {
			if (nextaclsuffix == NULL)
				nextaclsuffix = aclsuffixes;
			(void)snprintf(aclname, sizeof(aclname), "%s-%s",
			    gp->name, *nextaclsuffix);
			setastate(sp, ASTATE_SENTLISTACL);
			sfamily = family2str(ACLGROUP_FAMILY(gp));
			junoscriptsend(sp, js_listacl_fmt,
			    sfamily, aclname, sfamily);

			/* Advance to next suffix */
			++nextaclsuffix;
			if (*nextaclsuffix == NULL) {
				/* Done with this ACL */
				nextaclsuffix = NULL;
				++gp->sentlistacl;
			}
			return;
		}

	/*
	 * If we got here, we've listed all the ACLs
	 * Don't sort because Junos doesn't work that way
	 */
	++sp->listedallacls;
}

static void
junoscriptlistroute(struct state *sp, struct client *cl, struct req *rp,
    int nullzeroonly)
{
	clientsenderr(cl, rp, "junoscript doesn't support listroute");
}

static void
junoscriptlog(const char *what, const char *p)
{
	int len;
	const char *p2, *tssp;

	if (log_js.fn != NULL) {
		checklog(&log_js);
		tssp = tsstr();
		p2 = p;
		while (*p2 != '\0') {
			if (*p2 != '\r' && *p2 != '\n' && *p2 != '\0') {
				++p2;
				continue;
			}
			len = p2 - p;
			fprintf(log_js.f, "%s %s: %.*s\n", tssp, what, len, p);
			while (*p2 == '\r' || *p2 == '\n')
				++p2;
			p = p2;
		}
		if (*p != '\0')
			fprintf(log_js.f, "%s %s: %s\n", tssp, what, p);
		fflush(log_js.f);
	}
}

static void
junoscriptlogin(struct state *sp)
{
	int s;
	socklen_t salen;
	const char *errmsg;
	struct cf *cf;
	struct addr *a;
	struct addr addr;
	struct sockaddr_storage sa;
	struct sockaddr *sap;

	a = &addr;
	memset(a, 0, sizeof(*a));
	cf = sp->cf;
	errmsg = extractaddr(cf->c_router, NULL, a);
	if (errmsg != NULL) {
		lg(LOG_ERR, "junoscriptlogin: bad router address: %s: %s",
		    cf->c_router, errmsg);
		return;
	}

	/* Free ACL lists so we'll re-aquire them from the router */
	aclgroupsfree(sp);

	sap = (struct sockaddr *)&sa;
	salen = addr2sa(a, cf->c_portssh, sap);
	(void)snprintf(remote, sizeof(remote), "%s.%d", addr2str(a),
	    (int)cf->c_portssh);

	if (libssh2_init(0) < 0) {
		lg(LOG_ERR, "junoscriptlogin: libssh2_init failed");
		return;
	}

	s = socket(a->family, SOCK_STREAM, 0);
	if (s < 0) {
		lg(LOG_ERR, "junoscriptlogin: socket %s: %s",
		    remote, strerror(errno));
		return;
	}

	blocking(s, 0);
	errno = 0;
	if (connect(s, sap, salen) < 0) {
		if (errno != EINPROGRESS) {
			lg(LOG_ERR, "connect to %s: %s",
			    remote, strerror(errno));

			/* Clean up */
			(void)close(s);
			return;
		}
		lg(LOG_DEBUG, "connecting to %s", remote);
	} else {
		lg(LOG_ERR, "junoscriptlogin: Connect can't happen?: %s",
		    remote);
	}

	/* Enable child hook */
	sp->wfd = s;
	sp->f_childconnect = junoscriptconnect;
}

static const char *
junoscriptmapaclname(const char *groupname, struct acl *al)
{
	const char *p;
	static char aclname[64];

	p = val2str(acl2aclsuffix, al->type);
	if (p != NULL) {
		(void)snprintf(aclname, sizeof(aclname), "%s-%s", groupname, p);
		return (aclname);
	}
	return (groupname);
}

static void
junoscriptnullzero(struct state *sp, struct client *cl, struct req *rp)
{
	clientsenderr(cl, rp, "junoscript doesn't support null zero routes");
}

/* Qeueue config output to be sent after we have a private config */
static void
junoscriptqueue(struct state *sp, const char *fmt, ...)
{
	struct iobuf *ip;
	va_list ap;

	if (!sp->sentprivate) {
		junoscriptsend(sp, js_open_private_configuration);
		++sp->sentprivate;
	}

	ip = &sp->wbuf;
	va_start(ap, fmt);
	ioappendvfmt(ip, fmt, ap);
	va_end(ap);
}

static void
junoscriptquerynullzero(struct state *sp, struct client *cl, struct req *rp)
{
	clientsenderr(cl, rp, "junoscript doesn't support querynullzero");
}

/*
 * Read from a socket and append to an iobuf
 *
 * Return -1 if there was a fatal error, 0 on EOF or else 1
 */
int
junoscriptread(int fd, struct iobuf *ip)
{
	int rc, cc;
	char *cp;
	struct array *dp;

	dp = &ip->io_array;
	/* Loop until there's room for the EOS plus at least one new char */
	do {
		DYNARRAY(dp, ip->buf, dp->len + dp->inclen,
		    "junoscriptread buf");
		cp = ip->buf + IOBUF_LEN(ip);
		cc = IOBUF_SIZE(ip) - IOBUF_LEN(ip) - 1;
		rc = libssh2_channel_read(channel, cp, cc);
		if (rc < 0) {
			if (rc == LIBSSH2_ERROR_EAGAIN)
				return (1);
			if (rc == LIBSSH2_ERROR_CHANNEL_CLOSED)
				return (0);
			lg(LOG_ERR, "junoscriptread: libssh2_channel_read: %d",
			    rc);
			errno = EIO;
			return (-1);
		}
		if (rc > 0) {
			IOBUF_LEN(ip) += rc;
			ip->buf[IOBUF_LEN(ip)] = '\0';
			junoscriptlog("recv", cp);
		}
	} while (rc == cc);
	return (1);
}

static void
junoscriptrequest(struct state *sp, struct aclgroup *gp, struct req *rp)
{
	const char *sfamily, *sprotocol, *aclname, *fmt;

	/* Map the ACL group name to the ACL list name */
	aclname = junoscriptmapaclname(gp->name, &rp->acl);

	/* Family string for templates */
	sfamily = family2str(ACLGROUP_FAMILY(gp));

	/* Add or remove the ACL */
	switch (rp->type) {

	case REQ_DROP:
		junoscriptqueue(sp, js_drop_fmt,
		    sfamily, aclname, rp->acl.seq,
		    addr2str(&rp->acl.addr1), sfamily);
		break;

	case REQ_DROPUDPPORT:
	case REQ_DROPTCPPORT:
		if (rp->type == REQ_DROPUDPPORT)
			sprotocol = "udp";
		else if (rp->type == REQ_DROPTCPPORT)
			sprotocol = "tcp";
		else
			abort();
		if (ACLGROUP_ISIINET4(gp))
			fmt = js_dropport4_fmt;
		else if (ACLGROUP_ISIINET6(gp))
			fmt = js_dropport6_fmt;
		else
			abort();
		junoscriptqueue(sp, fmt,
		    aclname, rp->acl.seq, sprotocol, rp->acl.port1);
		break;

	case REQ_PERMITUDPDSTHOSTPORT:
	case REQ_PERMITTCPDSTHOSTPORT:
		if (rp->type == REQ_PERMITUDPDSTHOSTPORT)
			sprotocol = "udp";
		else if (rp->type == REQ_PERMITTCPDSTHOSTPORT)
			sprotocol = "tcp";
		else
			abort();
		if (ACLGROUP_ISIINET4(gp))
			fmt = js_permithostport4_fmt;
		else if (ACLGROUP_ISIINET6(gp))
			fmt = js_permithostport6_fmt;
		else
			abort();
		junoscriptqueue(sp, fmt, aclname, rp->acl.seq,
		    addr2str(&rp->acl.addr1), sprotocol, rp->acl.port1);
		break;

	case REQ_RESTORE:
	case REQ_RESTORETCPPORT:
	case REQ_RESTOREUDPPORT:
	case REQ_UNPERMITTCPDSTHOSTPORT:
	case REQ_UNPERMITUDPDSTHOSTPORT:
		junoscriptqueue(sp, js_delete_term_fmt,
		    sfamily, aclname, rp->acl.seq, sfamily);
		break;

	default:
		lg(LOG_ERR, "junoscriptrequest %s unsupported",
		    val2str(str2acl, rp->type));
		exit(EX_SOFTWARE);
	}

	/* Update request state */
	rp->state = RSTATE_CHILD;

	/* Wait for <load-configuration> before doing commit */
}

static void
junoscriptsend(struct state *sp, const char *fmt, ...)
{
	char *cp;
	size_t n;
	int rc;
	va_list ap;
	char buf[1024];

	va_start(ap, fmt);
	(void)vsnprintf(buf, sizeof(buf) - 1, fmt, ap);
	va_end(ap);

	/* Always append a newline */
	n = strlen(buf);
	cp = buf + n - 1;
	if (cp >= buf) {
		*++cp = '\n';
		*++cp = '\0';
		++n;
	}

	/* Send the line */
	rc = libssh2_channel_write(channel, buf, n);
	if (rc == LIBSSH2_ERROR_EAGAIN) {
		lg(LOG_ERR, "libssh2_channel_write: eagain");
		return;
	}
	if (rc < 0) {
		/* XXX kill off child here? */
		lg(LOG_ERR, "junoscriptsend: write: %s", strerror(errno));
		return;
	}

	if (libssh2_channel_eof(channel)) {
		/* XXX kill off child here? */
		lg(LOG_ERR, "junoscriptsend: EOF");
		return;
	}

	/* Log the line */
	junoscriptlog("send", buf);
}

static void
junoscriptsendattr(struct state *sp)
{
	++sp->sentattrs;
}

static void
junoscriptserverayt(struct state *sp)
{
	junoscriptsend(sp, js_uptime_cmd);
}

static void
junoscriptservercompact(struct state *sp)
{
	lg(LOG_DEBUG, "junoscriptservercompact: not implemented");
}

static void
junoscriptserversync(struct state *sp)
{
	junoscriptcommit(sp);
}

const char *
stripsuffix(const char *longaclname)
{
	char *cp;
	const char **pp;
	static char aclname[64];

	strncpy(aclname, longaclname, sizeof(aclname));
	aclname[sizeof(aclname) - 1] = '\0';
	cp = strrchr(aclname, '-');
	if (cp != NULL) {
		*cp++ = '\0';
		for (pp = aclsuffixes; *pp != NULL; ++pp)
			if (strcmp(cp, *pp) == 0)
				return (aclname);

	}
	return (longaclname);
}
