#!/usr/local/bin/python
# @(#) $Id: acld.py 1096 2015-10-03 02:25:21Z leres $ (LBL)
"""Communicate with acld"""

from __future__ import print_function

import errno
import os
import select
import socket
import time

DEFAULTINFLIGHT = 100
DEFAULTTIMEOUT = 90
DEFAULTADDR = '127.0.0.1'
DEFAULTPORT = 1965

FAILED_SUFFIX = '-failed'

FILE_CMDS = (
    'addwhitelist',
    'drop',
    'nonullzero',
    'nullzero',
    'query',
    'querynullzero',
    'querywhitelist',
    'remwhitelist',
    'restore',
)

CHECK_MAP = {
    'nullzero': 'querynullzero',
}

DRYRUN_MAP = {
    'nullzero': 'querynullzero',
    'nonullzero': 'querynullzero',
}

SINGLE_CMDS = (
    'state',
    'listnullzero',
    'listroute',
    'whitelist',
)

DOUBLE_CMDS = (
    'compact',
    'listacl',
)

TRIPLE_CMDS = (
    'blockhosthost',
    'restorehosthost',
    'dropudpport',
    'restoreudpport',
    'droptcpport',
    'restoretcpport',
    'permitudpdsthostport',
    'unpermitudpdsthostport',
    'permittcpdsthostport',
    'unpermittcpdsthostport',
)

class acld(object):
    """Communicate with acld"""
    def __init__(self, check=False, user=None, host=None, debug=False,
	dryrun=False, inflight=DEFAULTINFLIGHT, timeout=DEFAULTTIMEOUT,
	debugfile=None):
	"""Create an acld instance"""
	self.ibuf = ''
	if not user:
	    self.user = os.getlogin()
	else:
	    self.user = user
	if not host:
	    self.host = socket.gethostname()
	else:
	    self.host = host
	self.ident = '{%s@%s}' % (self.user, self.host)
	self.check = check
	self.debug = debug
	self.dryrun = dryrun
	self.inflight = inflight
	self.timeout = timeout
	self.debugfile = debugfile

	self.s = None
	self.cookie = 0
	self.inflightrequests = []

    def close(self):
	"""Close the acld socket"""
	if self.s:
	    # Send the exit
	    scookie = str(self.cookie)
	    self.cookie += 1
	    cmds = ['exit', scookie]
	    cmdline = ' '.join(cmds)
	    tup = [cmdline]
	    self.sendline(tup)
	    self.inflightrequests.append(cmds[0:2])
	    _tret, _last_tup, _payload = self.getresponse()

	    # Close the socket
	    self.s.close()
	    self.s = None

    def getacldline(self):
	"""Get a line from the acld socket and an optional error message"""
	while True:
	    tup = self.ibuf.split('\r\n', 1)
	    if len(tup) > 1:
		self.ibuf = tup[1]
		return (tup[0], None)

	    # Read more
	    rlist = [self.s]
	    wlist = []
	    xlist = []
	    timeout = self.timeout
	    try:
		rlist, wlist, xlist = select.select(rlist, wlist, xlist,
		    timeout)
	    except socket.error as e:
		errmsg = 'getacldline: select: %s' % e.strerror
		return (None, errmsg)

	    # Check for timeout
	    if not rlist and not wlist and not xlist:
		# XXX
		return (None,
		    'getacldline: acld read timeout (%.1f seconds)' % timeout)

	    if self.s in rlist:
		try:
		    buf = self.s.recv(1024)
		except socket.error as e:
		    errmsg = 'getacldline: recv: %s' % e.strerror
		    return (None, errmsg)
		cc = len(buf)
		if cc > 0:
		    # XXX
		    if self.debugfile:
			for line in buf.split('\r\n'):
			    print('<= %s' % line, file=self.debugfile)
		    self.ibuf += buf

    def getresponse(self):
	"""Reap response, return status, last tup and payload/error message"""
	ret = 0
	result = ''

	line, errmsg = self.getacldline()
	if errmsg:
	    return (1, [], errmsg)

	tup = line.split()
	n = len(tup)

	if n < 3 or n > 4:
	    errmsg = 'getresponse: Invalid response: %s' % line
	    return (1, [], errmsg)
	cookie = tup[1]

	if not self.inflightrequests:
	    return (1, [], 'Missing inflightrequests!')

	if cookie == self.inflightrequests[0][1]:
	    last_tup = self.inflightrequests.pop(0)
	else:
	    # Find the index of the request with this cookie
	    try:
		i = [x[1] for x in self.inflightrequests].index(cookie)
		last_tup = self.inflightrequests.pop(i)
	    except ValueError:
		last_tup = []

	    if not last_tup:
		errmsg = 'getresponse: Unexpected cookie: %s' % cookie
		if n == 4:
		    while True:
			line, errmsg2 = self.getacldline()
			if errmsg2:
			    errmsg = '\n'.join([errmsg, errmsg2])
			if line == '.':
			    break
		return (1, [], errmsg)

	what = last_tup[0]
	extended_what = tup[2]

	# Read optional payload
	payloads = []
	if n == 4:
	    if tup[3] != '-':
		errmsg = 'getresponse: Missing continuation: %s' % line
		return (1, [], errmsg)

	    result = 'SUCCESS'
	    while True:
		line, errmsg = self.getacldline()
		if line == '.':
		    break
		payloads.append(line)
		if errmsg:
		    payloads.append(line)

	if extended_what.endswith(FAILED_SUFFIX):
	    extended_what = extended_what[:-len(FAILED_SUFFIX)]
	    ret = 1
	    result = 'FAILED'

	if what != extended_what:
	    errmsg = 'getresponse: lose: %s vs. %s (%s)' % (what,
		extended_what, line)
	    return (1, [], errmsg)

	# Optionally report transaction
	if self.debug and what not in ('acld', 'exit'):
	    payloads.append(result)

	if payloads:
	    payload = '\n'.join(payloads)
	else:
	    payload = None

	return (ret, last_tup, payload)

    def open(self, addr=DEFAULTADDR, port=DEFAULTPORT):
	"""Open a socket to acld, return an optional error message"""
	tries = 3
	delay = 0.5
	while tries > 0:
	    errmsg = None
	    errnum = -1
	    tries -= 1
	    try:
		self.s = socket.create_connection((addr, port),
		    timeout=self.timeout)
	    except socket.error as e:
		detail = e.strerror
		errnum = e.errno
		if not detail:
		    detail = e.message
		if not detail:
		    detail = str(e)
		errmsg = 'create_connection: %s.%d: %s' % (addr, port, detail)

	    # Retry for some errors
	    # EPERM means the source port was in use
	    if errmsg and errnum in (errno.ECONNREFUSED, errno.EPERM):
		if tries > 0:
		    time.sleep(delay)
		    delay = delay * 2
		    continue
		return errmsg
	    break

	self.s.setblocking(0)

	self.cookie = 0
	self.inflightrequests = [['acld', '0']]
	ret, _last_tup, payload = self.getresponse()
	if ret:
	    errmsgs = ['bad greeting from acld']
	    if payload:
		errmsgs.append(payload)
	    self.s.close()
	    self.s = None
	    return ': '.join(errmsgs)

	return None

    def process(self, command, addrs, comments=None):
	"""Process list of addresses
	   If we're checking, always issue the check command first
	   When we get a check result, optionally issue the real command
	   Return a return status, a list of messages and an optional
	   list of error messages"""
	msgs = []
	errmsgs = []
	self.inflightrequests = []
	ret = 0

	# Handle check and dryrun modes
	if self.dryrun:
	    command2 = DRYRUN_MAP.get(command)
	elif self.check:
	    command2 = CHECK_MAP.get(command)
	else:
	    command2 = None

	if (self.dryrun or self.check) and not command2:
	    errmsg = "Can't find check command for %s" % command
	    errmsgs.append(errmsg)
	    return (1, msgs, errmsgs)

	# Loop on addrs from the file
	self.cookie = 0
	for addr in addrs:
	    scookie = str(self.cookie)
	    self.cookie += 1

	    # Build the next command
	    if command2:
		cmds = [command2, scookie, addr]
		cmdline = ' '.join(cmds)
		tup = [cmdline]
	    else:
		cmds = [command, scookie, addr, '-']
		cmdline = ' '.join(cmds)
		tup = [cmdline, self.ident]
		if comments:
		    tup.extend(comments)
		tup.append('.')

	    # Send the command
	    self.sendline(tup)
	    self.inflightrequests.append(cmds[0:3])

	    # Reap
	    if len(self.inflightrequests) >= self.inflight:
		tret, last_tup, payload = self.getresponse()

		# Is this a check result?
		# XXX
		if command2 and last_tup[0] == command2:
		    if self.dryrun:
			if payload and payload.startswith('N '):
			    fmt = 'Note: %s is already nullzero routed'
			    msg = fmt % last_tup[2]
			    msgs.append(msg)
			continue

		    if tret:
			scookie = str(self.cookie)
			self.cookie += 1
			addr = last_tup[2]
			cmds = [command, scookie, addr, '-']
			cmdline = ' '.join(cmds)

			# Send the command
			tup = [cmdline, self.ident]
			if comments:
			    tup.extend(comments)
			tup.append('.')
			self.sendline(tup)
			self.inflightrequests.append(cmds[0:3])
		    continue

		# Command result
		if tret:
		    ret = 1
		    # Append address
		    if len(last_tup) > 2:
			payload = ' '.join([payload, last_tup[2]])
		if payload:
		    msgs.append(payload)

	# Reap the stragglers
	while len(self.inflightrequests) > 0:
	    tret, last_tup, payload = self.getresponse()

	    # Is this a check result?
	    if command2 and last_tup[0] == command2:
		if self.dryrun:
		    if payload and payload.startswith('N '):
			msg = 'Note: %s is already nullzero routed' % last_tup[2]
			msgs.append(msg)
		    continue

		if tret:
		    scookie = str(self.cookie)
		    self.cookie += 1
		    addr = last_tup[2]
		    cmds = [command, scookie, addr, '-']
		    cmdline = ' '.join(cmds)

		    # Send the command
		    tup = [cmdline, self.ident]
		    if comments:
			tup.extend(comments)
		    tup.append('.')
		    self.sendline(tup)
		    self.inflightrequests.append(cmds[0:3])
		continue

	    # Command result
	    if tret:
		ret = 1
		# Append address
		if len(last_tup) > 2:
		    payload = ' '.join([payload, last_tup[2]])
	    if payload:
		msgs.append(payload)

	return (ret, msgs, errmsgs)

    def processone(self, args, comments=None):
	"""Process a single command
	   Return a return status, a list of messages and an optional
	   list of error messages"""
	msgs = []
	errmsgs = []
	self.cookie = 0
	ret = 0
	command = args[0]

	# Handle check and dryrun modes
	if self.dryrun:
	    command2 = DRYRUN_MAP.get(command)
	elif self.check:
	    command2 = CHECK_MAP.get(command)
	else:
	    command2 = None

	if (self.dryrun or self.check) and not command2:
	    errmsg = "Can't find check command for %s" % command
	    errmsgs.append(errmsg)
	    return (1, msgs, errmsgs)

	# Run the check command
	if command2:
	    # Build the command
	    scookie = str(self.cookie)
	    self.cookie += 1
	    cmds = [command2, scookie]
	    cmds += args[1:]
	    cmdline = ' '.join(cmds)

	    # Send the command
	    tup = [cmdline]
	    self.sendline(tup)
	    self.inflightrequests = [cmds]
	    tret, last_tup, payload = self.getresponse()
	    if self.dryrun:
		if command == 'nonullzero':
		    # Return same querynullzero result
		    return (tret, msgs, errmsgs)

		# nullzero cases
		if tret == 0:
		    # report the nullzero already exists
		    return (2, msgs, errmsgs)
		else:
		    # pretend we added a nullzero
		    return (0, msgs, errmsgs)

	    # Normal case
	    if tret == 0:
		msg = 'Skipping %s: (%s is yes)' % (' '.join(args), command)
		msgs.append(msg)
		return (2, msgs, errmsgs)

	# Build the command
	cmds = args[:]
	self.inflightrequests = [cmds]
	scookie = str(self.cookie)
	self.cookie += 1
	cmds.insert(1, scookie)
	cmds.append('-')
	cmdline = ' '.join(cmds)

	if self.dryrun:
	    msg = '+ %s' % cmdline
	    msgs.append(msg)
	else:
	    # Send the command
	    tup = [cmdline, self.ident]
	    if comments:
		tup.extend(comments)
	    tup.append('.')
	    self.sendline(tup)
	    ret, last_tup, payload = self.getresponse()
	    if payload:
		if ret and len(last_tup) > 2:
		    payload = ' '.join([payload, last_tup[2]])
		msgs.append(payload)
		# Return the special ""already exists" code
		if 'already' in payload:
		    ret = 2

	return (ret, msgs, errmsgs)

    def sendline(self, tup):
	"""Add \r\n's to each string in the list and concatenate
	   into a string"""
	if self.debugfile:
	    for line in tup:
		print('=> %s' % line, file=self.debugfile)
	buf = ''.join([x + '\r\n' for x in tup])
	self.s.send(buf)
