/*
 * Copyright (c) 2014
 * The Regents of the University of California. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the University nor the names of its contributors
 *       may be used to endorse or promote products derived from this software
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char rcsid[] =
    "@(#) $Id: rpc.c 1050 2014-09-03 20:34:16Z leres $ (LBL)";
#endif

#include <sys/types.h>

#include <stdio.h>
#include <string.h>
#include <sysexits.h>

#include "acld.h"
#include "junoscript.h"
#include "rpc.h"

/* rpc_parse_filter() flags */
#define PFLAG_ACCEPT	0x01
#define PFLAG_DISCARD	0x02
#define PFLAG_TCP	0x04
#define PFLAG_UDP	0x08

/* Forwards */
static xmlNode *rpc_child_element(xmlNode *);
static const char *rpc_get_content(xmlNode *);
static int rpc_ignore_warning(const char *);
static xmlNode *rpc_next_element(xmlNode *);
static xmlNode *rpc_parse_commit_results(struct state *, struct client *,
    struct req *, xmlNode *);
static const char *rpc_parse_filter(struct state *, xmlNode *);
static const char *rpc_parse_filter_term(xmlNode *, struct acl *, int *);
static xmlNode *rpc_parse_load_configuration(struct state *, struct client *,
    struct req *, xmlNode *);
static int rpc_parse_routing_engine(struct state *, struct client *,
    struct req *, xmlNode *);
static const char *rpc_parse_term_from(xmlNode *, struct acl *, int *);
static const char *rpc_parse_term_then(xmlNode *, struct acl *, int *);

/* Returns the child element node or NULL */
static xmlNode *
rpc_child_element(xmlNode *root)
{
	xmlNode *np;
	np = root->children;
	if (ISXML_TEXT(np)) {
		np = np->next;
		if (ISXML_ELEMENT(np))
			return (np);
	}
	return (NULL);
}

/* Return the text from this node's content or NULL */
static const char *
rpc_get_content(xmlNode *np)
{
	if (np != NULL) {
		np = np->children;
		if (ISXML_TEXT(np))
			return ((const char *)np->content);
	}
	return (NULL);
}

/* Return the text from the first <message> we find or NULL */
static const char *
rpc_get_message(xmlNode *np)
{
	np = rpc_child_element(np);
	for (; np != NULL; np = rpc_next_element(np)) {
		if (ISXML_ELEMENT_NAMED(np, "message")) {
			np = np->children;
			if (ISXML_TEXT(np))
				return ((const char *)np->content);
			return ("unknown");
		}
	}

	return (NULL);
}

static const char *ignore_warnings[] = {
	"uncommitted changes will be discarded on exit",
	"'synchronize' ignored in single routing engine chassis",
	NULL
};

/* Check for warning we ignore */
static int
rpc_ignore_warning(const char *str)
{
	const char **pp;

	for (pp = ignore_warnings; *pp != NULL; ++pp)
		if (strcmp(str, *pp) == 0)
			return (1);
	return (0);
}

/* Returns the next element node or NULL */
static xmlNode *
rpc_next_element(xmlNode *root)
{
	xmlNode *np;
	np = root->next;
	if (ISXML_TEXT(np)) {
		np = np->next;
		if (ISXML_ELEMENT(np))
			return (np);
	}
	return (NULL);
}

/* Return the number of commit-success tags seen */
static int
rpc_parse_routing_engine(struct state *sp, struct client *cl, struct req *rp,
    xmlNode *root)
{
	int success;
	xmlNode *np;
	const char *errmsg;

	success = 0;
	np = rpc_child_element(root);
	for (; np != NULL; np = rpc_next_element(np)) {
		if (ISXML_ELEMENT_NAMED(np, "error")) {
			errmsg = stripws(rpc_get_message(np));
			lg(LOG_DEBUG, "rpc_parse_routing_engine: error: \"%s\"",
			    pretty(errmsg));
			if (rp != NULL && rp->state == RSTATE_CHILD) {
				clientsenderr(cl, rp, "%s", pretty(errmsg));
				/* Log for NETS */
				nets_log(sp, cl, rp);
			}
			continue;
		}
		if (ISXML_ELEMENT_NAMED(np, "warning")) {
			errmsg = stripws(rpc_get_message(np));
			lg(LOG_DEBUG,
			    "rpc_parse_routing_engine: warning: \"%s\"",
			    pretty(errmsg));
			continue;
		}
		if (ISXML_ELEMENT_NAMED(np, "commit-success")) {
			++success;
			continue;
		}

		/* Ignore */
		if (ISXML_ELEMENT_NAMED(np, "name") ||
		    ISXML_ELEMENT_NAMED(np, "commit-check-success")) {
			continue;
		}

		lg(LOG_ERR, "rpc_parse_routing_engine: skipping \"%s\"",
		    (const char *)np->name);
	}
	return (success);
}

static xmlNode *
rpc_parse_commit_results(struct state *sp, struct client *cl, struct req *rp,
    xmlNode *root)
{
	int success;
	const char *errmsg;
	xmlNode *np;

	success = 0;
	np = rpc_child_element(root);
	for (; np != NULL; np = rpc_next_element(np)) {
		if (ISXML_ELEMENT_NAMED(np, "routing-engine")) {
			success += rpc_parse_routing_engine(sp, cl, rp, np);
			continue;
		}
		if (ISXML_ELEMENT_NAMED(np, "error")) {
			errmsg = stripws(rpc_get_message(np));
			lg(LOG_DEBUG, "rpc_parse_commit_results: error: \"%s\"",
			    pretty(errmsg));
			if (rp != NULL && rp->state == RSTATE_CHILD) {
				clientsenderr(cl, rp, "%s", errmsg);
				/* Log for NETS */
				nets_log(sp, cl, rp);
			}
			continue;
		}
		if (ISXML_ELEMENT_NAMED(np, "warning")) {
			errmsg = stripws(rpc_get_message(np));
			if (!rpc_ignore_warning(errmsg))
				lg(LOG_DEBUG,
				    "rpc_parse_commit_results: warning: \"%s\"",
				    pretty(errmsg));
			continue;
		}
		if (ISXML_ELEMENT_NAMED(np, "load-success")) {
			++success;
			continue;
		}

		lg(LOG_ERR, "rpc_parse_commit_results: skipping \"%s\"",
		    (const char *)np->name);
	}

	if (rp != NULL && rp->state == RSTATE_CHILD) {
		if (success > 0)
			clientsend(cl, rp);
		else
			clientsenderr(cl, rp,
			    "rpc_parse_commit_results: missing commit results");

		/* Log for NETS */
		nets_log(sp, cl, rp);
	}

	return (rpc_next_element(root));
}

const char *
rpc_parse_filter(struct state *sp, xmlNode *root)
{
	const char *p, *errmsg;
	xmlNode *np;
	struct cf *cf;
	const char *groupname, *aclname;
	struct addr *a;
	struct acl acl, *al;
	struct aclgroup *gp;
	int flags;
	int w;
	static char buf[64];
	enum acltype hosttype, nettype;
	enum acltype tcptype, udptype;
	enum acltype tcphosttype, tcpnettype, udphosttype, udpnettype;

	/* It's ok if the <configuration> is empty */
	np = rpc_child_element(root);
	if (np == NULL)
		return (NULL);

	if (!ISXML_ELEMENT_NAMED(np, "firewall"))
		return ("Can't find firewall node");
	np = rpc_child_element(np);

	if (!ISXML_ELEMENT_NAMED(np, "family"))
		return ("Can't find family node");
	np = rpc_child_element(np);

	if (!ISXML_ELEMENT_NAMED(np, "inet") &&
	    !ISXML_ELEMENT_NAMED(np, "inet6"))
		return ("family not inet or inet6");
	np = rpc_child_element(np);

	if (!ISXML_ELEMENT_NAMED(np, "filter"))
		return ("Can't find filter node");

	/* Step down to child and loop over terms */
	cf = sp->cf;
	al = &acl;
	memset(al, 0, sizeof(*al));
	errmsg = NULL;
	aclname = NULL;
	groupname = NULL;
	gp = NULL;
	np = rpc_child_element(np);
	for (; np != NULL; np = rpc_next_element(np)) {
		/* Initialize to parse next term */
		aclclean(al);
		memset(al, 0, sizeof(*al));
		al->seq = -1;
		flags = 0;

		/* Ignore */
		if (ISXML_ELEMENT_NAMED(np, "interface-specific"))
			continue;
		if (ISXML_ELEMENT_NAMED(np, "name")) {
			aclname = rpc_get_content(np);
			if (aclname != NULL) {
				/* Strip suffix to get ACL group name */
				groupname = stripsuffix(aclname);

				/* Find the aclgroup */
				gp = aclfindgroupbyname(cf, groupname);
				if (gp == NULL) {
					(void)snprintf(buf, sizeof(buf),
					    "Can't find ACL \"%s\"", groupname);
					return (buf);
				}
			}
			continue;
		}

		if (!ISXML_ELEMENT_NAMED(np, "term")) {
			lg(LOG_DEBUG,
			    "rpc_parse_filter: ignoring \"%s\" in filter %s",
			    (const char *)np->name, aclname);
			continue;
		}

		/* We should already havea parsed the filter name */
		if (gp == NULL)
			return ("rpc_parse_filter: Unknown ACL");

		/* Parse all term children nodes */
		p = rpc_parse_filter_term(np, al, &flags);
		if (p != NULL) {
			lg(LOG_ERR, "acl %s:%d %s", aclname, al->seq, p);
			continue;
		}

		/* Ignore this term if we didn't find a valid sequence number */
		if (al->seq < 0)
			continue;

		/* Add next acl: first do error checking */
		if ((flags & (PFLAG_ACCEPT | PFLAG_DISCARD)) == 0) {
			lg(LOG_ERR, "acl %s:%d Missing accept/discard",
			    aclname, al->seq);
			continue;
		}

		/* Shift single address to addr1 */
		if (al->addr1.family == 0 && al->addr2.family != 0) {
			al->addr1 = al->addr2;
			memset(&al->addr2, 0, sizeof(al->addr2));
		}

		/* Shift single port to port1 */
		if (al->port1 == NULL && al->port2 != NULL) {
			al->port1 = al->port2;
			al->port2 = NULL;
		}

		if (al->addr1.family != 0 && al->addr2.family == 0 &&
		    al->port1 == NULL && al->port2 == NULL) {
			/* host/net */
			a = &al->addr1;
			w = maskwidth(a);
			if ((flags & PFLAG_DISCARD) != 0) {
				hosttype = ATYPE_BLOCKHOST;
				nettype = ATYPE_BLOCKNET;
			} else {
				hosttype = ATYPE_PERMITHOST;
				nettype = ATYPE_PERMITNET;
			}

			if ((a->family == AF_INET && w == 32) ||
			    (a->family == AF_INET6 && w == 128))
				al->type = hosttype;
			else
				al->type = nettype;
		} else if (al->addr1.family == 0 && al->addr2.family == 0 &&
		    al->port1 != NULL && al->port2 == NULL &&
		    (flags & (PFLAG_TCP | PFLAG_UDP)) != 0) {
			/* udp/tcp port */
			if ((flags & PFLAG_DISCARD) != 0) {
				tcptype = ATYPE_BLOCKTCPPORT;
				udptype = ATYPE_BLOCKUDPPORT;
			} else {
				tcptype = ATYPE_PERMITTCPPORT;
				udptype = ATYPE_PERMITUDPPORT;
			}
			if ((flags & PFLAG_TCP) != 0)
				al->type = tcptype;
			else
				al->type = udptype;
		} else if (al->addr1.family != 0 && al->addr2.family == 0 &&
		    al->port1 != NULL && al->port2 == NULL &&
		    (flags & (PFLAG_TCP | PFLAG_UDP)) != 0) {
			/* udp/tcp dst port host/net */
			a = &al->addr1;
			w = maskwidth(a);
			if ((flags & PFLAG_DISCARD) != 0) {
				tcphosttype = ATYPE_BLOCKTCPDSTHOSTPORT;
				udphosttype = ATYPE_BLOCKUDPDSTHOSTPORT;
				tcpnettype = ATYPE_BLOCKTCPDSTNETPORT;
				udpnettype = ATYPE_BLOCKUDPDSTNETPORT;
			} else {
				tcphosttype = ATYPE_PERMITTCPDSTHOSTPORT;
				udphosttype = ATYPE_PERMITUDPDSTHOSTPORT;
				tcpnettype = ATYPE_PERMITTCPDSTNETPORT;
				udpnettype = ATYPE_PERMITUDPDSTNETPORT;
			}

			if ((flags & PFLAG_TCP) != 0) {
				hosttype = tcphosttype;
				nettype = tcpnettype;
			} else {
				hosttype = udphosttype;
				nettype = udpnettype;
			}

			if ((a->family == AF_INET && w == 32) ||
			    (a->family == AF_INET6 && w == 128))
				al->type = hosttype;
			else
				al->type = nettype;
		}

		if (al->type == ATYPE_UNKNOWN) {
			lg(LOG_ERR, "acl %s:%d Could not parse",
			    aclname, al->seq);
			continue;
		}

		if (IN_SEQRANGE(al->seq)) {
			if (gp->lastseq < al->seq)
				gp->lastseq = al->seq;
		} else if (IN_PERMITHOSTPORTSEQRANGE(al->seq)) {
			if (gp->lastpermithostportseq < al->seq)
				gp->lastpermithostportseq = al->seq;
		} else if (IN_PORTSEQRANGE(al->seq)) {
			if (gp->lastportseq < al->seq)
				gp->lastportseq = al->seq;
		} else {
			lg(LOG_ERR, "acl %s:%d outside seq ranges for %s",
			    aclname, al->seq, gp->name);
			continue;
		}

		acladdacl(sp, gp, al);
	}
	aclclean(al);
	return (errmsg);
}

static const char *
rpc_parse_filter_term(xmlNode *root, struct acl *al, int *flagsp)
{
	long lv;
	char *ep;
	xmlNode *np;
	const char *p, *errmsg;
	static char buf[64];

	np = rpc_child_element(root);
	for (; np != NULL; np = rpc_next_element(np)) {
		if (ISXML_ELEMENT_NAMED(np, "name")) {
			/* Parse sequence number */
			p = rpc_get_content(np);
			if (p == NULL) {
				lg(LOG_INFO, " Can't find term name");
				return (NULL);
			}
			lv = strtol(p, &ep, 10);
			if (*ep != '\0') {
				lg(LOG_INFO,
				    "skipping filter term named \"%s\"", p);
				return (NULL);
			}
			al->seq = lv;
			continue;
		}
		if (ISXML_ELEMENT_NAMED(np, "from")) {
			errmsg = rpc_parse_term_from(np, al, flagsp);
			if (errmsg != NULL)
				return (errmsg);
			continue;
		}
		if (ISXML_ELEMENT_NAMED(np, "then")) {
			errmsg = rpc_parse_term_then(np, al, flagsp);
			if (errmsg != NULL)
				return (errmsg);
			continue;
		}

		(void)snprintf(buf, sizeof(buf), "unknown filter term \"%s\"",
		    (const char *)np->name);
		return (buf);
	}
	return (NULL);
}

static xmlNode *
rpc_parse_load_configuration(struct state *sp, struct client *cl,
    struct req *rp, xmlNode *root)
{
	int success, error;
	xmlNode *np;
	const char *errmsg;

	success = 0;
	error = 0;
	np = rpc_child_element(root);
	for (; np != NULL; np = rpc_next_element(np)) {
		if (ISXML_ELEMENT_NAMED(np, "error")) {
			errmsg = stripws(rpc_get_message(np));
			lg(LOG_DEBUG,
			    "rpc_parse_load_configuration: error: \"%s\"",
			    pretty(errmsg));
			if (rp != NULL && rp->state == RSTATE_CHILD) {
				clientsenderr(cl, rp, "%s", errmsg);
				/* Log for NETS */
				nets_log(sp, cl, rp);
			}
			++error;
			continue;
		}
		if (ISXML_ELEMENT_NAMED(np, "warning")) {
			errmsg = stripws(rpc_get_message(np));
			lg(LOG_DEBUG,
			    "rpc_parse_load_configuration: warning: \"%s\"",
			    pretty(errmsg));
			if (strstr(errmsg, "statement not found"))
				++error;
			continue;
		}
		if (ISXML_ELEMENT_NAMED(np, "load-success")) {
			/* Now we can commit the configuration change */
			++success;
			continue;
		}

		/* Ignore */
		/* XXX shouldn't we care? */
		if (ISXML_ELEMENT_NAMED(np, "load-error-count"))
			continue;

		lg(LOG_ERR,
		    "rpc_parse_load_configuration: skipping \"%s\"",
		    (const char *)np->name);
	}
	if (success && !error)
		junoscriptcommit(sp);
	return (rpc_next_element(root));
}

static const char *
rpc_parse_term_from(xmlNode *root, struct acl *al, int *flagsp)
{
	struct addr *a;
	xmlNode *np, *np2;
	const char *p, *errmsg;
	static char buf[64];

	np = rpc_child_element(root);
	for (; np != NULL; np = rpc_next_element(np)) {
		if (ISXML_ELEMENT_NAMED(np, "address") ||
		    ISXML_ELEMENT_NAMED(np, "source-address") ||
		    ISXML_ELEMENT_NAMED(np, "destination-address")) {
			np2 = rpc_child_element(np);
			if (!ISXML_ELEMENT_NAMED(np2, "name"))
				return ("missing address name");
			p = rpc_get_content(np2);
			/* "address" and "source-address" in the 1st addr */
			if (ISXML_ELEMENT_NAMED(np, "destination-address"))
				a = &al->addr2;
			else
				a = &al->addr1;
			errmsg = extractaddr(p, NULL, a);
			if (errmsg != NULL)
				return (errmsg);
			continue;
		}
		if (ISXML_ELEMENT_NAMED(np, "protocol") ||
		    ISXML_ELEMENT_NAMED(np, "next-header")) {
			p = rpc_get_content(np);
			if (strcmp(p, "tcp") == 0)
				*flagsp |= PFLAG_TCP;
			else if (strcmp(p, "udp") == 0)
				*flagsp |= PFLAG_UDP;
			else {
				(void)snprintf(buf, sizeof(buf),
				    "unknown protocol \"%s\"", p);
				return (buf);
			}
			continue;
		}
		if (ISXML_ELEMENT_NAMED(np, "destination-port")) {
			p = rpc_get_content(np);
			if (al->port1 != NULL)
				free(al->port1);
			al->port1 = strsave(p);
			continue;
		}

		(void)snprintf(buf, sizeof(buf), "unknown from \"%s\"",
		    (const char *)np->name);
		return (buf);
	}
	return (NULL);
}

static const char *
rpc_parse_term_then(xmlNode *root, struct acl *al, int *flagsp)
{
	xmlNode *np;
	static char buf[64];

	np = rpc_child_element(root);
	for (; np != NULL; np = rpc_next_element(np)) {
		if (ISXML_ELEMENT_NAMED(np, "discard")) {
			*flagsp |= PFLAG_DISCARD;
			continue;
		}
		if (ISXML_ELEMENT_NAMED(np, "accept")) {
			*flagsp |= PFLAG_ACCEPT;
			continue;
		}

		(void)snprintf(buf, sizeof(buf), "unknown then \"%s\"",
		    (const char *)np->name);
		return (buf);
	}
	return (NULL);
}

static xmlNode *
rpc_parse_uptime(xmlNode *root)
{
	xmlNode *np;
	const char *p;
	const char *s_time;
	const char *s_uptime;
	const char *s_users;
	const char *s_load1;
	const char *s_load5;
	const char *s_load15;

	s_time = "?";
	s_uptime = "?";
	s_users = "?";
	s_load1 = "?";
	s_load5 = "?";
	s_load15 = "?";

	/* We only want to look at uptime-information children */
	np = rpc_child_element(root);
	for (; np != NULL; np = rpc_next_element(np)) {
		if (ISXML_ELEMENT_NAMED(np, "uptime-information")) {
			np = rpc_child_element(np);
			break;
		}
	}

	for (; np != NULL; np = rpc_next_element(np)) {
		if (ISXML_ELEMENT_NAMED(np, "date-time")) {
			s_time = rpc_get_content(np);
			continue;
		}
		if (ISXML_ELEMENT_NAMED(np, "up-time")) {
			s_uptime = rpc_get_content(np);
			continue;
		}
		if (ISXML_ELEMENT_NAMED(np,
		    "active-user-count")) {
			s_users = rpc_get_content(np);
			continue;
		}
		if (ISXML_ELEMENT_NAMED(np, "load-average-1")) {
			s_load1 = rpc_get_content(np);
			continue;
		}
		if (ISXML_ELEMENT_NAMED(np, "load-average-5")) {
			s_load5 = rpc_get_content(np);
			continue;
		}
		if (ISXML_ELEMENT_NAMED(np, "load-average-15")) {
			s_load15 = rpc_get_content(np);
			continue;
		}

		lg(LOG_ERR, "rpc_parse_uptime: skipping \"%s\"",
		    (const char *)np->name);
	}

	if (strcmp(s_users, "1") == 0)
		p = "";
	else
		p = "s";

	lg(LOG_INFO, "%s up %s, %s user%s, load averages: %s, %s, %s",
	    s_time, s_uptime, s_users, p, s_load1, s_load5, s_load15);
	return (rpc_next_element(root));
}

const char *
rpc_process(struct state *sp, struct client *cl, struct req *rp, xmlNode *root)
{
	xmlNode *np;
	const char *p, *errmsg;
	static char buf[132];

	errmsg = NULL;
	if (!ISXML_ELEMENT_NAMED(root, "rpc-reply")) {
		(void)snprintf(buf, sizeof(buf),
		    "rpc_process: unexpected xml %s", (const char *)root->name);
		return (buf);
	}

	/* Process reply */
	np = rpc_child_element(root);
	while (np != NULL) {
		if (ISXML_ELEMENT_NAMED(np, "configuration")) {
			errmsg = rpc_parse_filter(sp, np);
			np = rpc_next_element(np);
			continue;
		}
		if (ISXML_ELEMENT_NAMED(np, "error")) {
			errmsg = stripws(rpc_get_message(np));
			lg(LOG_DEBUG, "rpc_process: error: \"%s\"",
			    pretty(errmsg));
			if (rp != NULL && rp->state == RSTATE_CHILD) {
				clientsenderr(cl, rp, "%s", errmsg);
				/* Log for NETS */
				nets_log(sp, cl, rp);
			}
			np = rpc_next_element(np);
			continue;
		}
		if (ISXML_ELEMENT_NAMED(np, "warning")) {
			p = stripws(rpc_get_message(np));
			if (!rpc_ignore_warning(p)) {
				lg(LOG_DEBUG, "rpc_process: warning: \"%s\"",
				    pretty(errmsg));
				errmsg = p;
			}
			np = rpc_next_element(np);
			continue;
		}
		if (ISXML_ELEMENT_NAMED(np, "load-configuration-results")) {
			np = rpc_parse_load_configuration(sp, cl, rp, np);
			continue;
		}
		if (ISXML_ELEMENT_NAMED(np, "system-uptime-information")) {
			np = rpc_parse_uptime(np);
			continue;
		}
		if (ISXML_ELEMENT_NAMED(np, "commit-results")) {
			/* Exit private configuration mode */
			junoscriptexitprivate(sp);
			np = rpc_parse_commit_results(sp, cl, rp, np);
			continue;
		}

		lg(LOG_ERR, "rpc_process: skipping \"%s\"",
		    (const char *)np->name);
		np = rpc_next_element(np);
	}

	return (errmsg);
}
